#include <iostream.h>
#include <qobjcoll.h>
#include "geomgr.h"

// Forward ref
static void layoutWidget(LayoutInfo *item, QSize &oldsize, QSize &newsize);

/*!
  Constructs a new geometry manager which will be attached to \a parent.
  The geometry manager installs itself as the event filter for its parent so
  it can intercept any resizeEvents. The manager is not yet active.
*/

GeoMgr::GeoMgr(QWidget *parent, const char *name)
: QObject(parent, name)
{
  _target = parent;
  _list.setAutoDelete(TRUE);
  if(parent)
    parent->installEventFilter(this);
  _active = FALSE;
}

/*!
  Destroys a geometry manager.
*/

GeoMgr::~GeoMgr()
{
  _list.clear();
}

/*!
  Initializes the geometry manager. Every time new widgets are added to
  the parent, this member should be called. If it is not called, the
  newly added widgets will not be managed. Each time init() is called, the
  layout information of each subwidget will be reset to the default values
  of \e NO_GRAV, \e NO_GRAV and \e RESIZE_ALL.

  Init() gets a list of child widgets. The GeoMgr manages only the
  subwidgets in the first degree. You can, however, install additional
  Geometry managers for each of the subwidgets. For each subwidget, a
  \e LayoutInfo structure is created with default values (linear scaling).
  The entries can be modified.

   \sa \e setNWGrav(), \e setSEGrav(), \e setLayoutInfo()
*/

void GeoMgr::init()
{
  _list.clear();
  QObjectList *list = _target->queryList("QWidget", 0, TRUE, FALSE);
  QObjectListIt it(*list);
  while(it.current())
  {
    QWidget *target = (QWidget *)(it.current());

    // Take care not to include internal subwidgets like scrollbars on
    // QListBox and QMultiLineEdit. Also, QMenuBar manages its own geometry.

    QString ptype = target->parent()->className();
    QString ctype = target->className();
    if(ptype != "QListbox" && ptype != "QMultiLineEdit"
       && ctype != "QMenuBar")
    {
      LayoutInfo *entry = new LayoutInfo;
      entry->target = target;
      entry->nw_grav = NO_GRAV;
      entry->se_grav = NO_GRAV;
      entry->mode = RESIZE_ALL;
      _list.append(entry);
    }
    ++it;
  }
  delete list;
  _active = TRUE;
}

/*!
  Sets the gravity mode of the Northwest control point of the widget pointed
  to by \a target. If \a target cannot be found, 0 is returned.
*/

int GeoMgr::setNWGrav(QWidget *target, GravityMode nw)
{
  LayoutInfo *info = findEntry(target);
  if(info)
  {
    info->nw_grav = nw;
    return 1;
  }
  else
    return 0;
}

/*!
  Sets the gravity mode of the Southeast control point of the widget pointed
  to by \a target. If \a target cannot be found, 0 is returned.
*/

int GeoMgr::setSEGrav(QWidget *target, GravityMode se)
{
  LayoutInfo *info = findEntry(target);
  if(info)
  {
    info->se_grav = se;
    return 1;
  }
  else
    return 0;
}

/*!
  Sets the resize mode of the widget pointed to by \a target. If \a target
  cannot be found, 0 is returned.
*/

int GeoMgr::setResizeMode(QWidget *target, ResizeMode mode)
{
  LayoutInfo *info = findEntry(target);
  if(info)
  {
    info->mode = mode;
    return 1;
  }
  else
    return 0;
}

/*!
  Sets the layout info of the widget pointed to by \a target. If \a target
  cannot be found, 0 is returned.
*/

int GeoMgr::setLayoutInfo(QWidget *target, GravityMode nw, GravityMode se,
                   ResizeMode mode)
{
  LayoutInfo *info = findEntry(target);
  if(info)
  {
    info->nw_grav = nw;
    info->se_grav = se;
    info->mode = mode;
    return 1; //Success
  }
  else
    return 0; //Failure
}

/*!
  Returns the layout info of the widget pointed to by \a target. If \a target
  cannot be found, 0 is returned. The gravity modes and resize mode are
  returned in the arguments.
*/

int GeoMgr::getLayoutInfo(QWidget *target, GravityMode &nw, GravityMode &se,
                  ResizeMode &mode)
{
  LayoutInfo *info = findEntry(target);
  if(info)
  {
    nw = info->nw_grav;
    se = info->se_grav;
    mode = info->mode;
    return 1; //Success
  }
  else
    return 0; //Failure
}

//
// private members
//

/*!
  Processes events sent to the parent. Only resizeEvents are processed by
  the geometry manager. All other events are dispatched to the parent
  normally.
*/

bool GeoMgr::eventFilter(QObject *, QEvent *ev)
{
  if(ev->type() == Event_Resize)
  {
    doLayout((QResizeEvent *)ev);
    return TRUE;
  }
  return FALSE;
}

/*!
  Goes through the list of subwidgets under geometry control and calls
  layoutWidget(LayoutInfo *, QSize &, QSize &) for each of them.
*/

void GeoMgr::doLayout(QResizeEvent *ev)
{
  // Do nothing as long as init() has not been called
  if(! _active)
    return;

  // Check if our parent is visible. If not, bail out
  QWidget *par = (QWidget *)parent();
  if(! par->isVisible())
    return;

  QSize oldsize = ev->oldSize();
  QSize newsize = ev->size();

  LayoutInfo *curr = _list.first();
  while(curr)
  {
    layoutWidget(curr, oldsize, newsize);
    curr = _list.next();
  }
}

/*!
  Calculate the new position and size of a single widget based on the
  gravity modes, resize mode and the old and new size of the parent.
*/

void GeoMgr::layoutWidget(LayoutInfo *item, QSize &oldsize, QSize &newsize)
{
  // Calculate scaling parameters
  double xscale = (double)newsize.width()/oldsize.width();
  double yscale = (double)newsize.height()/oldsize.height();
  int deltax = newsize.width() - oldsize.width();
  int deltay = newsize.height() - oldsize.height();

  // Calculate the new position of the Northwest control point
  QPoint Pnw = item->target->pos();
  GravityMode nw_grav = item->nw_grav;

  // Calculate the new x-position
  if(nw_grav == N_GRAV || nw_grav == S_GRAV || nw_grav == NO_GRAV)
    Pnw.setX((int)(Pnw.x()*xscale));
  else if(nw_grav == NE_GRAV || nw_grav == E_GRAV || nw_grav == SE_GRAV)
    Pnw.setX(Pnw.x()+deltax);
  else if(nw_grav == SW_GRAV || nw_grav == W_GRAV || nw_grav == NW_GRAV)
    Pnw.setX(Pnw.x());

  // Calculate the new y-position
  if(nw_grav == E_GRAV || nw_grav == W_GRAV || nw_grav == NO_GRAV)
    Pnw.setY((int)(Pnw.y()*yscale));
  else if(nw_grav == SE_GRAV || nw_grav == S_GRAV || nw_grav == SW_GRAV)
    Pnw.setY(Pnw.y()+deltay);
  else if(nw_grav == N_GRAV || nw_grav == NE_GRAV || nw_grav == NW_GRAV)
    Pnw.setY(Pnw.y());

  // Calculate the new position of the Southeast control point
  QPoint Pse = item->target->pos()
               + QPoint(item->target->width(), item->target->height());
  GravityMode se_grav = item->se_grav;

  // Calculate the new x-position
  if(se_grav == N_GRAV || se_grav == S_GRAV || se_grav == NO_GRAV)
    Pse.setX((int)(Pse.x()*xscale));
  else if(se_grav == NE_GRAV || se_grav == E_GRAV || se_grav == SE_GRAV)
    Pse.setX(Pse.x()+deltax);
  else if(se_grav == SW_GRAV || se_grav == W_GRAV || se_grav == NW_GRAV)
    Pse.setX(Pse.x());

  // Calculate the new y-position
  if(se_grav == E_GRAV || se_grav == W_GRAV || se_grav == NO_GRAV)
    Pse.setY((int)(Pse.y()*yscale));
  else if(se_grav == SE_GRAV || se_grav == S_GRAV || se_grav == SW_GRAV)
    Pse.setY(Pse.y()+deltay);
  else if(se_grav == N_GRAV || se_grav == NE_GRAV || se_grav == NW_GRAV)
    Pse.setY(Pse.y());

  ResizeMode rmode = item->mode;
  QWidget *widget = item->target;

  // Calculate the new center of gravity
  QPoint Pgrav = (Pnw + Pse)/2;

  // Check what resize mode this widget resize it
  if(rmode == RESIZE_ALL)
    widget->resize(Pse.x()-Pnw.x(), Pse.y()-Pnw.y() );
  else if(rmode == RESIZE_X)
    widget->resize(Pse.x()-Pnw.x(), item->target->height() );
  else if(rmode == RESIZE_Y)
    widget->resize(item->target->width(), Pse.y()-Pnw.y() );
  // RESIZE_NONE doesn't require resize

  widget->move(Pgrav.x() - widget->width()/2,
               Pgrav.y() - widget->height()/2);
}

/*!
  Looks up a widget in the list of subwidgets under geometry control. If
  \a target cannot be found, a null pointer is returned.
*/

LayoutInfo *GeoMgr::findEntry(QWidget *target)
{
  LayoutInfo *curr = _list.first();
  while(curr)
  {
    if(curr->target == target)
      return curr;
    curr = _list.next();
  }
  return (LayoutInfo *)0;
}
