#ifndef geomgr_included
#define geomgr_included

#include <qwidget.h>
#include <qlist.h>

/*!
  \class GravityMode
  \brief Enumerate for gravity modes
*/

enum GravityMode
{
  NO_GRAV,
  N_GRAV, NE_GRAV,
  E_GRAV, SE_GRAV,
  S_GRAV, SW_GRAV,
  W_GRAV, NW_GRAV
};

/*!
  \class ResizeMode
  \brief Enumerate for resize modes
*/

enum ResizeMode
{
  RESIZE_ALL,
  RESIZE_X,
  RESIZE_Y,
  RESIZE_NONE
};

/*!
  \class LayoutInfo
  \brief Structure used as an child widget list entry
*/

typedef struct 
{
  QWidget *target;
  GravityMode nw_grav;
  GravityMode se_grav;
  ResizeMode mode;
}
LayoutInfo;

/*!
  \class GeoMgr geomgr.h
  \brief Provides a geometry manager for subwidgets

  The GeoMgr class provides a geometry manager based on the gravity
  principle also found in the X-window system. Rather than using grid-like
  layout and stretch parameters, it uses gravity settings for two
  opposite corners of each subwidget called the northwest and southeast
  control points. The top of the subwidget is defined as the north side of
  the widget.

  For each subwidget the gravity mode of the two control points can be set.
  The modes have the following effect:

  Gravity mode
  <ul>
  <li> \e N_GRAV        Distance to north edge remains constant
  <li> \e NE_GRAV       Distance to northeast corner remains constant
  <li> \e E_GRAV        Distance to east edge remains constant
  <li> \e SE_GRAV       Distance to southeast corner remains constant
  <li> \e S_GRAV        Distance to south edge remains constant
  <li> \e SW_GRAV       Distance to southwest corner remain constant
  <li> \e W_GRAV        Distance to west edget remains constant
  <li> \e NW_GRAV       Distance to northwest corner remains constant
  <li> \e NO_GRAV       Point position is linearly scaled to match the new size
  </ul>

  If, for example, for a subwidget both control points are set to \e North
  gravity, the subwidget will keep both its northwest and its southeast
  corners at a constant distance to the north side of its parent. If the
  parent is made higher, the subwidget will not move or resize. If the
  parent is made wider, the subwidget will be linearly moved and scaled.
  Only gravity is not enough to accomplish all desired effects. You might
  wish to move a subwidget like a pushbutton in such a way that it keeps the
  same relative position (the center for instance) in its parent. This would
  be accomplished by setting the gravity of both control points to \e None.
  Unfortunately, this would resize the pushbutton also. This may be
  undesired. To accomodate this, one more attribute can be set for the
  layout information of a subwidget. This is the ResizeMode. It can have
  the following values: \e RESIZE_ALL, \e RESIZE_X, \e RESIZE_Y,
  \e RESIZE_NONE.

  The geometry manager needs to be activated by calling the init() member
  function after it has been constructed. This creates a list of \e LayoutInfo
  entries. One entry is created for each direct descendant of the parent.
  The gravity mode of both control points is set to \e NO_GRAV and the
  resize mode is set to \e RESIZE_ALL. In effect, everything scales
  linearly. Only \e direct descendants are included in the list of
  subwidgets that are managed by the geometry manager. If you have
  subwidgets which themselves have children, the you should add a second
  geometry manager to that subwidget (see example).

  \warning
    If you add new widgets to the parent of the geometry manager, you should
    call init() again. Otherwise, these widgets will not be included in the
    geometry management.
  \warning
    You should only change gravity modes and/or resize modes \e after
    calling init().
  
  Code example:
  \code
    #include <geomgr.h>

    class MyClass: public QWidget
    {
      public:
        ....
      private:
        GeoMgr *mainMgr;
        GeoMgr *subMgr;
        QFrame *backFrame;
        QPushButton *quitBut, *frameBut;
    }

    MyClass::MyClass(QWidget *parent, const char *name)
    : QWidget(parent, name)
    {
      backFrame = new QFrame(this);
      ....
      quitBut = new QPushButton(this);
      ....
      frameBut = new QPushButton(backFrame);
      mainMgr = new GeoMgr(this);         // manages children of MyClass
      mainMgr->init();
      subMgr = new GeoMgr(backFrame);     // manages the children of the frame
      subMgr->init();
    }
  \endcode
*/

class GeoMgr: public QObject
{
  Q_OBJECT
  public:
    GeoMgr(QWidget *parent=0, const char *name=0);
    ~GeoMgr();
    void init();
    int setNWGrav(QWidget *target, GravityMode nw);
    int setSEGrav(QWidget *target, GravityMode se);
    int setResizeMode(QWidget *target, ResizeMode mode);
    int setLayoutInfo(QWidget *target, GravityMode nw, GravityMode se,
                      ResizeMode mode);
    int getLayoutInfo(QWidget *target, GravityMode &nw, GravityMode &se,
                      ResizeMode &mode);
  private:
    bool eventFilter(QObject *, QEvent *);
    void doLayout(QResizeEvent *);
    void layoutWidget(LayoutInfo *, QSize &, QSize &);
    LayoutInfo *findEntry(QWidget *target);
    //! The target of which the child geometry should be managed
    QWidget *_target;
    //! A list of LayouInfo of child widgets under geometry management
    QList<LayoutInfo> _list;
    //! Activity flag
    bool _active;
};

#endif
