#include <iostream.h>
#include <math.h>

#include "placehld.h"

/*!
  Creates a new SizeSlab with a size of 6 x 6 pixels and a black background.
*/

SizeSlab::SizeSlab(QWidget *parent, char *name)
: QWidget(parent, name)
{
  resize(6, 6);
  setBackgroundColor(black);
  _dragging = FALSE;
}

/*!
  Destoy a SizeSlab.
*/

SizeSlab::~SizeSlab()
{
  //cerr << "deleting SizeSlab\n";
}

/*!
  \fn void SizeSlab::displace(const QPoint &delta)

  This signal is emitted when the mouse is dragged while it was pressed down
  in this SizeSlab. It reports the displacement from the point where the
  mouse button was pressed.
*/

/*!
  \fn void SizeSlab::released()

  This signal is emitted when the mouse button is released after dragging.
*/
  
/*!
  Handles mouse press events. The _dragging flag is set and the point where
  the mouse button was pressed down is remembered.
*/

void SizeSlab::mousePressEvent(QMouseEvent *ev)
{
  _dragging = TRUE;
  _startPoint = ev->pos();
}

/*!
  Handles mouse release events. The _dragging flag is reset and the
  released() signal is emitted.
*/

void SizeSlab::mouseReleaseEvent(QMouseEvent *)
{
  emit released();
  _dragging = FALSE;
}

/*!
  Handles mouse move events. If the user is dragging the mouse, the
  displacement relative to the point where the mouse was pressed is emitted
  through the displace(const QPoint &) signal.
*/

void SizeSlab::mouseMoveEvent(QMouseEvent *ev)
{
  if(_dragging)
  {
    QPoint delta = ev->pos() - _startPoint;
    emit displace(delta);
  }
}

/*!
  Creates a new PlaceHolder. You should not supply 0 as the parent. This
  would make the SizeSlabs used by the PlaceHolder windows of their own,
  resulting in decorations being added by the window manager. The
  PlaceHolder is hidden by default until its target is set.
*/

PlaceHolder::PlaceHolder(QWidget *parent, char *name)
: QObject(parent, name)
{
  _gridx = 10;
  _gridy = 10;

  _topLeft = new SizeSlab(parent, "_topLeft");
  _topRight = new SizeSlab(parent, "_topRight");
  _bottomLeft = new SizeSlab(parent, "_bottomLeft");
  _bottomRight = new SizeSlab(parent, "_bottomRight");
  _center = new SizeSlab(parent, "_center");
  _target = 0;

  _topLeft->setCursor(SizeFDiagCursor);
  _topRight->setCursor(SizeBDiagCursor);
  _bottomLeft->setCursor(SizeBDiagCursor);
  _bottomRight->setCursor(SizeFDiagCursor);
  _center->setCursor(SizeAllCursor);

  connect(_topLeft, SIGNAL(displace(const QPoint &)),
           SLOT(topLeftDisplace(const QPoint &)) );
  connect(_topRight, SIGNAL(displace(const QPoint &)),
           SLOT(topRightDisplace(const QPoint &)) );
  connect(_bottomLeft, SIGNAL(displace(const QPoint &)),
           SLOT(bottomLeftDisplace(const QPoint &)) );
  connect(_bottomRight, SIGNAL(displace(const QPoint &)),
           SLOT(bottomRightDisplace(const QPoint &)) );
  connect(_center, SIGNAL(displace(const QPoint &)),
           SLOT(centerDisplace(const QPoint &)) );
  connect(_topLeft, SIGNAL(released()), SLOT(displaceDone()) );
  connect(_topRight, SIGNAL(released()), SLOT(displaceDone()) );
  connect(_bottomLeft, SIGNAL(released()), SLOT(displaceDone()) );
  connect(_bottomRight, SIGNAL(released()), SLOT(displaceDone()) );
  connect(_center, SIGNAL(released()), SLOT(displaceDone()) );
  hide();
}

/*!
  Deletes the PlaceHolder.
*/

PlaceHolder::~PlaceHolder()
{
  //cerr << "deleting PlaceHolder\n";
}

/*!
  Hides the five SizeSlab widgets, making the PlaceHolder invisible.
*/

void PlaceHolder::hide()
{
  _topLeft->hide();
  _topRight->hide();
  _bottomLeft->hide();
  _bottomRight->hide();
  _center->hide();
}

/*!
  Shows all five SizeSlab widgets if a valid target is set. Before they are
  shown, all five SizeSlabs are raised to make sure they are the topmost
  widgets on the list of children of the target.
*/

void PlaceHolder::show()
{
  if(_target != 0)
  {
    _topLeft->raise();
    _topRight->raise();
    _bottomLeft->raise();
    _bottomRight->raise();
    _center->raise();
    _topLeft->show();
    _topRight->show();
    _bottomLeft->show();
    _bottomRight->show();
    _center->show();
  }
}

/*!
  Returns whether the PlaceHolder is visible.
*/

bool PlaceHolder::isVisible()
{
  return _center->isVisible();
}

/*!
  Sets the target of the PlaceHolder. The SizeSlab widgets are placed in the
  correct position and the PlaceHolder is made visible.
*/

void PlaceHolder::setTarget(QWidget *target)
{
  _target = target;
  placeSlabs();
  show();
}

/*!
  Returns the target widget of the PlaceHolder.
*/

QWidget *PlaceHolder::getTarget()
{
  return _target;
}

/*!
  \fn void PlaceHolder::setGrid(int x, int y)

  Set the grid to \a x times \y pixels. The grid is used to snap the target
  widget to a grid on the mutual parent widget.
*/

// SLOTS

/*!
  The top left SizeSlab was moved, resize/move the whole widget to match the
  displacement.
*/

void PlaceHolder::topLeftDisplace(const QPoint & delta)
{
  _target->move(_target->pos() + delta);
  _target->resize(_target->size() - QSize(delta.x(), delta.y()) );
  _topLeft->move(_topLeft->pos()+delta);
  _topRight->move(_topRight->x(), _topRight->y()+delta.y());
  _bottomLeft->move(_bottomLeft->x()+delta.x(), _bottomLeft->y());
  _center->move(_center->pos() + delta/2);
}

/*!
  The top right SizeSlab was moved, resize/move the whole widget to match the
  displacement.
*/

void PlaceHolder::topRightDisplace(const QPoint & delta)
{
  _target->move(_target->pos() + QPoint(0, delta.y()) );
  _target->resize(_target->size() + QSize(delta.x(), -delta.y()) );
  _topLeft->move(_topLeft->x(), _topLeft->y()+delta.y());
  _topRight->move(_topRight->pos() + delta);
  _bottomRight->move(_bottomRight->x()+delta.x(), _bottomRight->y());
  _center->move(_center->pos() + delta/2);
}

/*!
  The bottom left SizeSlab was moved, resize/move the whole widget to match
  the displacement.
*/

void PlaceHolder::bottomLeftDisplace(const QPoint & delta)
{
  _target->move(_target->pos() + QPoint(delta.x(), 0) );
  _target->resize(_target->size() + QSize(-delta.x(), delta.y()) );
  _topLeft->move(_topLeft->x()+delta.x(), _topLeft->y());
  _bottomLeft->move(_bottomLeft->pos() + delta);
  _bottomRight->move(_bottomRight->x(), _bottomRight->y()+delta.y());
  _center->move(_center->pos() + delta/2);
}

/*!
  The bottom right SizeSlab was moved, resize/move the whole widget to match
  the displacement.
*/

void PlaceHolder::bottomRightDisplace(const QPoint & delta)
{
  _target->resize(_target->size() + QSize(delta.x(), delta.y()) );
  _topRight->move(_topRight->x()+delta.x(), _topRight->y());
  _bottomLeft->move(_bottomLeft->x(), _bottomLeft->y()+delta.y());
  _bottomRight->move(_bottomRight->pos() + delta);
  _center->move(_center->pos() + delta/2);
}

/*!
  The center SizeSlab was moved, move the whole widget to match the
  displacement.
*/

void PlaceHolder::centerDisplace(const QPoint & delta)
{
  _target->move(_target->pos() + delta);
  _topLeft->move(_topLeft->pos() + delta);
  _topRight->move(_topRight->pos() + delta);
  _bottomLeft->move(_bottomLeft->pos() + delta);
  _bottomRight->move(_bottomRight->pos() + delta);
  _center->move(_center->pos() + delta);
}

/*!
  Snap the resized/moved widget back to the nearest grid points defined by
  the _gridx and _gridy members.
*/

void PlaceHolder::displaceDone()
{
  // Make sure the target component is placed and sized to the grid
  int x = _target->x();
  int y = _target->y();
  int w = _target->width();
  int h = _target->height();

  x = _gridx*(int)rint(x/_gridx);
  y = _gridy*(int)rint(y/_gridy);
  w = _gridx*(int)rint(w/_gridx);
  h = _gridy*(int)rint(h/_gridy);
  _target->setGeometry(x, y, w, h);
  placeSlabs();
}

//
// Private members
//

/*!
  Place the SizeSlab widgets on the correct position on the target widget.
  This is a bit complicated, as our parent may not be the same as the parent
  of the target widget.
*/

void PlaceHolder::placeSlabs()
{
  if(! _target)
    return;

  QPoint slabSize(_center->width(), _center->height());

  QWidget *targetParent = (QWidget *)(_target->parent());
  QPoint targetPos = _target->pos();
  QPoint targetGlobalPos;
  if(targetParent)
    targetGlobalPos = targetParent->mapToGlobal(targetPos);
  else
    targetGlobalPos = targetPos;
  QWidget *myParent = (QWidget *)parent();
  QPoint targetRealPos;
  if(myParent)
    targetRealPos = myParent->mapFromGlobal(targetGlobalPos);
  else
    targetRealPos = targetGlobalPos;

  QPoint tPos(targetRealPos);
  QPoint tSize(_target->width(), _target->height());

  _topLeft->move(tPos);
  _topRight->move(tPos.x() + tSize.x() - slabSize.x(), tPos.y());
  _bottomLeft->move(tPos.x(), tPos.y() + tSize.y() - slabSize.y());
  _bottomRight->move(tPos +tSize - slabSize);
  _center->move(tPos + (tSize - slabSize)/2);
}
