#ifndef PlaceHolder_included
#define PlaceHolder_included

#include <qobject.h>
#include <qwidget.h>

/*!
  \class SizeSlab placehld.h
  \brief Slab used as a handle by PlaceHolder class to resize or move a Widget

  A SizeSlab emits signals when the mouse button is pressed on it and it is
  dragged. The displacement from the point where the mouse was pressed is
  emitted with each mouseMoveEvent which is received. When the mouse button
  is released, another signal is sent.
*/

class SizeSlab: public QWidget
{
  Q_OBJECT
  public:
    SizeSlab(QWidget *parent=0, char *name=0);
    ~SizeSlab();
  signals:
    void displace(const QPoint &delta);
    void released();
  protected:
    void mousePressEvent(QMouseEvent *);
    void mouseReleaseEvent(QMouseEvent *);
    void mouseMoveEvent(QMouseEvent *);
  private:
    //! Flag that indicated mouse dragging
    bool _dragging;
    //! The start point of mouse dragging
    QPoint _startPoint;
};

/*!
  \class PlaceHolder
  \brief Places itself on top of any target Widget and provides five handles
         to move and resize the target.

  The PlaceHolder class provides you with a means to resize or move any
  widget. By using the \e setTarget() member, you can specify the widget it
  should attach to. Typically you should create the PlaceHolder with the
  same parent as the target widget. The PlaceHolder will automatically place
  itself at the top of the child list of the mutual parent by calling
  raise()to make sure that no other widgets with the same parent obstruct
  it. As soon as you set the target to a valid widget, the PlaceHolder will
  make itself visible. It will place four SizeSlab widgets at the corners of
  the target widget and one at the center of the target.

  The PlaceHolder has a grid setting. This grid setting is used to round the
  size and position of the target widget to the nearest gridpoints. The
  default value for the grid is 10x10 pixels. It can be changed with the \e
  setGrid() member function. The grid is not used during drag operations. As
  soon as the mouse button is released after dragging, the target widget is
  snapped to the grid.
  
  Note:
  The sizeslabs are
  created by the PlaceHolder, but they are created as children of the target
  widget. Therefore, they are not deleted in the destructor of the
  PlaceHolder. This is not a memory leak!

  \warning
  Make sure the target set for the PlaceHolder is visible before you attach
  it. Otherwise it is undefined what will happen! Also, make sure you hide()
  the PlaceHolder before you delete its target widget!
*/

class PlaceHolder: public QObject
{
  Q_OBJECT
  public:
    PlaceHolder(QWidget *parent=0, char *name=0);
    ~PlaceHolder();
    void hide();
    void show();
    bool isVisible();
    void setTarget(QWidget *target);
    QWidget *getTarget();
    void setGrid(int x, int y) {_gridx = x; _gridy = y;}
  public slots:
    void topLeftDisplace(const QPoint &delta);
    void topRightDisplace(const QPoint &delta);
    void bottomLeftDisplace(const QPoint &delta);
    void bottomRightDisplace(const QPoint &delta);
    void centerDisplace(const QPoint &delta);
    void displaceDone();
  protected:
    void placeSlabs();
  private:
    //! The top left SizeSlab
    SizeSlab *_topLeft;
    //! The top right SizeSlab
    SizeSlab *_topRight;
    //! The bottom left SizeSlab
    SizeSlab *_bottomLeft;
    //! The bottom right SizeSlab
    SizeSlab *_bottomRight;
    //! The center SizeSlab
    SizeSlab *_center;

    //! The target of the PlaceHolder
    QWidget *_target;
    //! The grid value for the x-direction
    int _gridx;
    //! The grid value for the y-direction
    int _gridy;
};

#endif
