#ifndef StorageModule_included
#define StorageModule_included

#include <stdio.h>
#include <fstream.h>

#include <qobject.h>
#include <qwidget.h>
#include <qstring.h>

#include "shared.h"

/*!
  \class ParserCode
  \brief Enumeration of lexical analyser return codes
*/

enum ParserCode
{
  EndOfInput,
  WidgetStart,
  PropertyReady,
  MemberFuncStart,
  MemberVarStart,
  IncludeStart,
  OpenBracket,
  CloseBracket,
  ParseError
};

/*!
  \class StorageModule storemdl.h
  \brief Module which takes care of serialization and exporting of WidgetInfo
         objects.

  The StorageModule takes care of all serialization tasks in Ebuilder. It
  supports three public members: read, write and export. The write and
  export members take a WidgetInfo object as argument and write the content
  to disk in either a proprietary format or C++ code. The read member can
  reconstruct a WidgetInfo object from a file previously written to disk by
  the write member function.

  This class uses a lexical analyser generated with flex to analyse the
  input.
*/

class StorageModule: public QObject
{
  Q_OBJECT
  public:
    StorageModule(QObject *parent=0, const char *name=0);
    void export(WidgetInfo *info, const char *basename);
    WidgetInfo *read(const char *fname);
    void write(WidgetInfo *info, const char *fname);
  private:
    QWidget *addWidget(QWidget *parent, const QString &type
                       , const QString &name);
    int setWidgetProperty(QWidget *target, const QString &key
                           , const QString &value);
    int setMemberFuncProperty(MemberFunction *target, const QString &key
                           , const QString &value);
    int setMemberVarProperty(MemberVariable *target, const QString &key
                           , const QString &value);
    int setIncludeProperty(IncludeFile *target, const QString &key
                           , const QString &value);
    void push(QWidget *val) {stack[sp++] = val;}
    QWidget *pop() {return stack[--sp];}
    void writeWidgetProps(QWidget *target, ofstream &fp);
    void writeMemberFuncs(const QList<MemberFunction> &list, ofstream &fp);
    void writeMemberVars(const QList<MemberVariable> &list, ofstream &fp);
    void writeIncludes(const QList<IncludeFile> &list, ofstream &fp);
    void exportWidgetProps(QWidget *target, ofstream &fp);
    int checkProperty(QWidget *target, const char *property);

    //! Variable for lexical analyser holding widget type
    QString type;
    //! Variable for lexical analyser holding widget name
    QString name;
    //! Variable for lexical analyser holding property name
    QString key;
    //! Variable for lexical analyser holding property value
    QString value;
    //! Variable for lexical analyser holding the current line number
    int lineNumber;
    //! Variable for lexical analyser holding the current recursion depth
    int level;
    //! Variable for lexical analyser holding the current stack depth
    int sp;
    //! Variable for lexical analyser forming a widget stack for recursion
    QWidget *stack[100];

    friend int yylex();
};

extern StorageModule *parseObj;

#endif
