#include <iostream.h>
#include <qpainter.h>
#include <qobjcoll.h>
#include "tree.h"

/*!
  Constructs a TreeNode
*/

TreeNode::TreeNode(const char *txt, QPixmap *pic, void *data)
{
  _txt = txt;
  _pic = pic;
  _data = data;
  _level = 0;
  _highlighted = FALSE;
  _expanded = TRUE;
  _rect = QRect(0, 0, 0, 0);
}

/*!
  Destroys a TreeNode
*/

TreeNode::~TreeNode()
{
}

/*!
  \fn void TreeNode::setHighlighted(bool val)

  Set the \e highlighted attribute of a TreeNode.
*/

/*!
  \fn bool TreeNode::highlighted()

  Returns the \e highlighted attribute of a TreeNode.
*/

/*!
  \fn void TreeNode::setExpanded(bool val)

  Set the \e expanded attribute of a TreeNode.
*/

/*!
  \fn bool TreeNode::expanded()

  Returns the \e expanded attribute of a TreeNode.
*/

/*!
  \fn void TreeNode::setText(const char *txt)

  Sets the text that will be drawn in the TreeNode.
*/

/*!
  \fn const char *TreeNode::text()

  Returns the text that is drawn in the TreeNode.
*/

/*!
  \fn void TreeNode::setPixmap(QPixmap *pic)

  Sets the pixmap that will be drawn in the TreeNode. If \a pic is 0, no
  pixmap will be drawn.
*/

/*!
  \fn QPixmap *TreeNode::pixmap()

  Returns the pixmap that is use to draw the TreeNode.
*/

/*!
  \fn void TreeNode::setData(void *data)

  Sets the data associated with the TreeNode.
*/

/*!
  \fn void *TreeNode::data()

  Returns the data associated with the TreeNode.
*/

/*!
  \fn void TreeNode::setLevel(int val)

  Set the indentation level of the TreeNode.
*/

/*!
  \fn int TreeNode::level()

  Returns the indentation level of the TreeNode.
*/

/*!
  \fn QRect TreeNode::rect()

  Returns the rectangle occupied by the treeNode. The coordinates are
  relative to the upper left corner of the screen area.
*/

/*!
  Draw a treenode on paintdevice \a dev at coordinates \a x, \a y. The node
  is drawn on a dark background if the node is highlighted.
*/

void TreeNode::draw(QPaintDevice *dev, int x, int y)
{
  QFontMetrics fm(QFont::defaultFont());

  int width = 0;
  if(_txt)
    width = fm.width(_txt) + 2;
  if(_pic)
    width += TREE_NODE_HEIGHT;
  _rect = QRect(x, y, width, TREE_NODE_HEIGHT);

  // Draw this treeNode
  QPainter painter;
  painter.begin(dev);

  if(_pic)
  {
    painter.drawPixmap(x+1, y+1, *_pic);
    x += TREE_NODE_HEIGHT;
  }
  else
   x++;

  if(_highlighted)
  {
    painter.setPen(white);
    if(_txt)
      painter.fillRect(x, y+(TREE_NODE_HEIGHT-fm.height())/2,
                       fm.width(_txt)+2, fm.height(), darkBlue);
    else
      painter.fillRect(x, y+(TREE_NODE_HEIGHT-fm.height())/2,
                       50, fm.height(), darkBlue);
  }

  painter.drawText(x, y, fm.width(_txt)+2, TREE_NODE_HEIGHT,
                   AlignHCenter | AlignVCenter, _txt);

  painter.end();
}

/*!
  Construct a TreeWidget. The \a parent and \a name arguments are sent to
  the QWidget constructor.
*/

TreeWidget::TreeWidget(QWidget *parent, const char *name)
: QFrame(parent, name)
{
  setFrameStyle(QFrame::Panel | QFrame::Sunken);
  setBackgroundColor(white);

  _xBar = new QScrollBar(QScrollBar::Horizontal, this);
  _xBar->setGeometry(0, height()-15, width()-15, 15);
  _xBar->setRange(0, 0);
  _xBar->setValue(0);
  _xBar->setSteps(TREE_NODE_HEIGHT, width()-25);
  _yBar = new QScrollBar(QScrollBar::Vertical, this);
  _yBar->setGeometry(width()-15, 0, 15, height()-15);
  _yBar->setRange(0, 0);
  _yBar->setValue(0);
  _yBar->setSteps(TREE_NODE_HEIGHT, height()-25);
  connect(_xBar, SIGNAL(valueChanged(int)), SLOT(doXBar(int)) );
  connect(_yBar, SIGNAL(valueChanged(int)), SLOT(doYBar(int)) );

  _rootNode = 0;
  _currNode = 0;
  _nodeList.setAutoDelete(TRUE);
  _offset.setX(0);
  _offset.setY(0);
  _canvas = new QPixmap(TREE_NODE_HEIGHT, TREE_NODE_HEIGHT);

  _nodePopup = 0;
  _backPopup = 0;
}

/*!
  Destroy the TreeWidget.
*/

TreeWidget::~TreeWidget()
{
}

/*!
  Set the current node of the TreeWidget to the node identified by \a id.
  The new current node will be highlighted. If \a id is null, nothing will
  be selected.
*/

void TreeWidget::setCurrentNode(TreeNode *id)
{
  if(_currNode)
    _currNode->setHighlighted(FALSE);
  _currNode = id;
  if(id)
    _currNode->setHighlighted(TRUE);
  update();
}

/*!
  Returns a TreeNode pointer to the root node of the TreeWidget.
*/

TreeNode *TreeWidget::rootNode()
{
  return _rootNode;
}

/*!
  Returns the TreeNode pointer of the currently selected node of the
  TreeWidget. If no node is selected, null is returned.
*/

TreeNode *TreeWidget::currentNode()
{
  return _currNode;
}

/*!
  Returns a list of TreeNode pointers of all nodes in the tree. 
  Example:
  \code
    NodeList *list = myTree->nodes();
    NodeListIt it(*list);
    TreeNode *curr;
    for(; it.current(); ++it)
    {
      ....
    }
    delete list;
  \endcode
  
  \warning
     Delete the list as soon as you are done. It is allocated specifically
     for you. If you don't delete it, you'l have a memory leak on your hands!
*/

NodeList *TreeWidget::nodes()
{
  NodeList *ret = new NodeList;
  ret->setAutoDelete(FALSE);

  TreeNode *curr = _nodeList.first();
  while(curr)
  {
    ret->append(curr);
    curr = _nodeList.next();
  }
  return ret;
}

/*!
  Adds a new node to the tree using \a txt as the node text and \a pic as
  the pixmap. The \a data parameter is an opaque container which
  is cast to (void *). This way you can store proprietary data along with
  the new node. You can later use the data() member function of the TreeNode
  to retrieve this data. The \a parent parameter is used to locate the parent
  for the new tree node. If \a parent is 0, the new node will be
  used as the root node of the tree. Note that you can only have one root
  node. If a root node is present, you will need to delete it first using
  rmNode().

  A pointer to the newly created TreeNode is returned. This pointer can
  be used as a reference later if you wish to retrieve information about the
  node.

  \warning
  The TreeWidget retains ownership of \e all the TreeNode-s in the tree. Do
  not \e ever attempt to delete a TreeNode!
*/

TreeNode *TreeWidget::addNode(const char *txt, QPixmap *pic, void *data,
                               TreeNode *parent)
{
  TreeNode *node;

  if(parent)
  {
    if(_nodeList.findRef(parent) >= 0)
    {
      int level = _nodeList.current()->level();
      while(_nodeList.next())
      {
        if(_nodeList.current()->level() <= level)
          break;
      }
      if(_nodeList.at() < 0)
        _nodeList.last();
      else
        _nodeList.prev();

      node = new TreeNode(txt, pic, data);
      node->setLevel(1 + parent->level());

      _nodeList.insert(_nodeList.at()+1, node);
    }
    else
    {
      cerr << "*** addNode: Unable to find parent node\n";
      return (TreeNode *)0;
    }
  }
  else
  {
    if(_rootNode)
    {
      cerr << "*** A rootnode allready exists, there can be only one!\n";
      return (TreeNode *)0;
    }
    else
    {
      node  = new TreeNode(txt, pic, data);
      _rootNode = node;
      _nodeList.insert(0, _rootNode);
    }
  }

  update();
  return node;
}

/*!
  Adds a new node to the tree. The node should be created by the user. It is
  placed in the tree as a child of the treenode pointed to by \a parent. The
  TreeWidget assumes ownership of the TreeNode after it has been added to
  the tree in this way. However, it does not own the data and pixmap members
  of the TreeNode. Freeing the storage related to those is up to the user.
*/
  
void TreeWidget::addNode(TreeNode *node, TreeNode *parent)
{
  if(parent)
  {
    if(_nodeList.findRef(parent) >= 0)
    {
      int level = _nodeList.current()->level();
      while(_nodeList.next())
      {
        if(_nodeList.current()->level() <= level)
          break;
      }
      if(_nodeList.at() < 0)
        _nodeList.last();
      else
        _nodeList.prev();

      node->setLevel(1 + parent->level());

      _nodeList.insert(_nodeList.at()+1, node);
    }
    else
    {
      cerr << "*** addNode: Unable to find parent node\n";
      return;
    }
  }
  else
  {
    if(_rootNode)
    {
      cerr << "*** A rootnode allready exists, there can be only one!\n";
      return;
    }
    else
    {
      _rootNode = node;
      node->setLevel(0);
      _nodeList.insert(0, _rootNode);
    }
  }
  update();
}

/*!
  Remove the tree node identified by \a target. The corresponding node and
  all it's children will be deleted.

  \warning
  If the TreeNode contains a pixmap or data, neither of these will be
  deleted! Whoever create the node with addNode() retains ownership of
  this memory and has the reponsibility to delete it.
*/

void TreeWidget::rmNode(TreeNode *target)
{
  if(_nodeList.findRef(target) < 0)
  {
    cerr << "*** rmNode: Unable to find node\n";
    return;
  }

  if(target == _currNode)
    _currNode = 0;

  int level = target->level();
  int idx = _nodeList.at();
  target = _nodeList.next();

  while(target)
  {
    if(target->level() > level)
    {
      if(_nodeList.at() == (int)_nodeList.count()-1)
      {
        _nodeList.remove();
        break;
      }
      else
        _nodeList.remove();
    }
    else
      break;
    target = _nodeList.current();
  }
  _nodeList.remove(idx);
  update();
}

/*!
  Clear the whole TreeWidget. All nodes in the tree are deleted, including
  the root node.
  \warning
  If any of the TreeNodes contain pixmaps or data, neither of these will be
  deleted! Whoever create the node with addNode() retains ownership of
  this memory and has the reponsibility to delete it.
*/

void TreeWidget::clear()
{
  _nodeList.clear();
  _rootNode = 0;
  _currNode = 0;
  update();
}

/*!
  \fn void TreeWidget::setNodePopup(QPopupMenu *menu)

  Specifies the popup menu that should be shown when the user right clicks
  on a selected node. By default, nothing hapens in this case. You can use
  this function to determine the operations that can be done on a node
  Example:
  \code
    MyClass::MyClass(QWidget *parent, const char *name)
    : QWidget(parent, name)
    {
      TreeWidget *tree = new TreeWidget(this);
      tree->setGeometry(...);
      ....
      QPopupMenu *nodePopup = new QPopupMenu(0);
      nodePopup->insertItem("&Delete", this, SLOT(deleteNode()) );
      nodePopup->insertItem("&Copy", this, SLOT(copyNode()) );
      nodePopup->insertItem("&Edit", this, SLOT(editNode()) );
      ....
      tree->setNodePopup(nodePopup);
    }
  \endcode

  The \e deleteNode() SLOT could look like this:
  \code
    void MyClass::deleteNode(void)
    {
      ....
      TreeNode *curr = tree->currentNode();
      tree->rmNode(curr);
      ....
    }
  \endcode
*/

/*!
  \fn void TreeWidget::setBackPopup(QPopupMenu *menu)

  Specifies the popup menu that should be shown when the user right clicks
  on a the background of the TreeWidget. By default, nothing hapens in this
  case. You can use this function to determine the operations that are
  available for the whole tree object.
  Example:
  \code
    MyClass::MyClass(QWidget *parent, const char *name)
    : QWidget(parent, name)
    {
      TreeWidget *tree = new TreeWidget(this);
      tree->setGeometry(...);
      ....
      QPopupMenu *backPopup = new QPopupMenu(0);
      backPopup->insertItem("&Clear", this, SLOT(clearTree()) );
      backPopup->insertItem("&Quit", this, SLOT(quit()) );
      ....
      tree->setBackPopup(backPopup);
    }
  \endcode
*/

/*!
  \fn void TreeWidget::doXBar(int val)

  Handle movement of the x-scrollbar. Scroll the treewidget left and right.
*/

/*!
  \fn void TreeWidget::doYBar(int val)

  Handle movement of the y-scrollbar. Scroll the treewidget up and down.
*/

/*!
  \fn void TreeWidget::highlighted()

  This signal is emitted when a new item has been highlighted by selecting
  it with the left mouse button.
*/

/*!
  \fn void TreeWidget::selected()

  This signal is emitted when a new item is selected by double-clicking a
  tree node with the left mouse button.
*/

/*!
  \fn void TreeWidget::nothingSelected()

  This signal is emitted when all selections are cancelled by clicking
  the left mouse button in the background of the TreeWidget.
*/

/*!
  Handles mouse click events. If the left mouse button is pressed inside the
  area used to draw a node, that node is selected. If the left mouse button
  is pressed in the background area which is not used to draw nodes on, all
  nodes are deselected and \e nothingSelected() is emitted.
  If the right mouse button is pressed on a node, the nodePopup menu is
  shown. If it is pressed in the background area, the backPopup menu is
  shown.
*/

void TreeWidget::mousePressEvent(QMouseEvent *ev)
{

  // Find out if a node was selected with this press
  TreeNode *target = 0;
  QPoint pos = ev->pos() + _offset - QPoint(10, 10);
  TreeNode *curr = _nodeList.first();
  while(curr)
  {
    if(curr->rect().contains(pos))
    {
      target = curr;
      break;
    }
    curr = _nodeList.next();
  }

  if(ev->button() == LeftButton)
  {
    if(_currNode)
      _currNode->setHighlighted(FALSE);
    if(target)
    {
      _currNode = target;
      _currNode->setHighlighted(TRUE);
      emit highlighted();
    }
    else
    {
      _currNode = 0;
      emit nothingSelected();
    }
    update();
  }
  else if(ev->button() == RightButton)
  {
    if(target)
    {
      if(_nodePopup)
      {
        if(_currNode)
        {
          _nodePopup->move(mapToGlobal(ev->pos()));
          _nodePopup->show();
	}
      }
    }
    else
    {
      if(_backPopup)
      {
        _backPopup->move(mapToGlobal(ev->pos()));
        _backPopup->show();
      }
    }
  }
}

/*!
  Handles mouse doubleclick events
*/

void TreeWidget::mouseDoubleClickEvent(QMouseEvent *)
{
  if(_currNode)
    emit selected();
}

/*!
  Paints the buffer containing the pixmap pixture on the screen.
*/

void TreeWidget::paintEvent(QPaintEvent *ev)
{
  QFrame::paintEvent(ev);
  QPainter painter;
  painter.begin(this);
  painter.setClipRect(10, 10, width()-25, height()-25);
  painter.drawPixmap(10, 10, *_canvas, _offset.x(), _offset.y());
  painter.end();
}

/*!
  Handles resize events. Calls \e update() to ensure that the background
  buffer pixmap is appropriately resized and updated. This function also
  updates the range and value of the x- and y scrollbars
*/

void TreeWidget::resizeEvent(QResizeEvent *)
{
  _xBar->setGeometry(0, height()-15, width()-15, 15);
  _yBar->setGeometry(width()-15, 0, 15, height()-15);
  update();
  _xBar->setSteps(TREE_NODE_HEIGHT, width()-25);
  _yBar->setSteps(TREE_NODE_HEIGHT, height()-25);
}

/*!
  Paints the tree hierarchy into the background buffer.
  \warning
  This function \ul does not call \e update(). The content of the background
  buffer is not drawn on the screen!
*/

void TreeWidget::draw()
{
  QPainter painter;
  int maxWidth=TREE_NODE_HEIGHT, maxHeight=TREE_NODE_HEIGHT;
  _canvas->fill();

  // Draw the nodes
  int xpos = 0, ypos = 0;
  TreeNode *curr = _nodeList.first();
  while(curr)
  {
    xpos = TREE_NODE_HEIGHT * curr->level();
    curr->draw(_canvas, xpos, ypos);

    // Remember the size of the canvas
    QRect rect = curr->rect();
    if(rect.right() > maxWidth)
      maxWidth = rect.right();
    if(rect.bottom() > maxHeight)
      maxHeight = rect.bottom();

    ypos += TREE_NODE_HEIGHT;
    curr = _nodeList.next();
  }

  // Check if the canvas is too big
  if((maxWidth < _canvas->width()) || (maxHeight < _canvas->height()))
  {
    _canvas->resize(maxWidth, maxHeight);
  }

  // Check if the canvas is big enough
  if((maxWidth > _canvas->width()) || (maxHeight > _canvas->height()))
  {
    _canvas->resize(maxWidth, maxHeight);
    _canvas->fill();

    // Redraw the nodes on the new canvas;
    xpos = 0; ypos = 0;
    curr = _nodeList.first();
    while(curr)
    {
      xpos = TREE_NODE_HEIGHT * curr->level();
      curr->draw(_canvas, xpos, ypos);
      ypos += TREE_NODE_HEIGHT;
      curr = _nodeList.next();
    }
  }

  // Draw the lines connecting the nodes
  curr = _nodeList.first();
  int first=1;
  xpos = 0; ypos = 0;
  painter.begin(_canvas);
  QPen mypen(gray, 2);
  painter.setPen(mypen);
  while(curr)
  {
    if(!first)
    {
      xpos = TREE_NODE_HEIGHT * curr->level();
      painter.drawLine(xpos-TREE_NODE_HEIGHT/2, ypos+TREE_NODE_HEIGHT/2,
                       xpos, ypos+TREE_NODE_HEIGHT/2);
    }
    ypos += TREE_NODE_HEIGHT;
    curr = _nodeList.next();
    first=0;
  }

  int length = _nodeList.count();
  for(int node = 0; node < length; node++)
  {
    TreeNode *srcnode = _nodeList.at(node);
    int last=node;

    for(int cmp = node+1; cmp < length; cmp++)
    {
      TreeNode *cmpnode = _nodeList.at(cmp);
      if(srcnode->level() == cmpnode->level())
        break;
      if(srcnode->level() == cmpnode->level()-1)
        last = cmp;
    }
    if(last != node)
    {
      int x = TREE_NODE_HEIGHT * srcnode->level() + TREE_NODE_HEIGHT/2;
      int y1 = TREE_NODE_HEIGHT * (node+1);
      int y2 = TREE_NODE_HEIGHT * last + TREE_NODE_HEIGHT/2;
      painter.drawLine(x, y1, x, y2);
    }
  }
  painter.end();

  // Calibrate the size of the scrollbars
  if(_canvas->height() > height()-35)
    _yBar->setRange(0, _canvas->height() - height() + 35);
  else
    _yBar->setRange(0, 0);

  if(_canvas->width() > width()-35)
    _xBar->setRange(0, _canvas->width() - width() + 35);
  else
    _xBar->setRange(0, 0);
}

/*!
  Force an update of the information in the tree. Call this function if you
  have updated the content of any of the TreeNode-s. If you don't call this
  function then you will have to wait until the next forced repaint.
*/

void TreeWidget::update()
{
  draw();
  QWidget::update();
}
