#ifndef TreeWidget_included
#define TreeWidget_included

#include <iostream.h>
#include <qframe.h>
#include <qlist.h>
#include <qscrbar.h>
#include <qpixmap.h>
#include <qpopmenu.h>

#define TREE_PIXMAP_SIZE 16
#define TREE_NODE_HEIGHT 18

class TreeNode;
class TreeWidget;

/*!
  \class
     TreeNode tree.h
  \brief
     This is a class for internal use. It is used by the TreeWidget
     class to hold data for the nodes in the tree

  This class is used by the TreeWidget to implement the tree nodes. It can
  be used to get information on the data, text and pixmap information of
  this node.

  The treenode provides some functionality which is only available to the
  TreeWidget class. This is necessary to shield users of the TreeNode from
  functionality they do not require.

  \sa TreeWidget
*/

class TreeNode
{
  public:
    TreeNode(const char *txt, QPixmap *pic=0, void *data=0);
    ~TreeNode();
    void setHighlighted(bool val) {_highlighted = val;}
    bool highlighted() {return _highlighted;}
    void setExpanded(bool val) {_expanded = val;}
    bool expanded() {return _expanded;}
    void setText(const char *txt) {_txt = txt;}
    const char *text() {return _txt;}
    void setPixmap(QPixmap *pic) {_pic = pic;}
    QPixmap *pixmap() {return _pic;}
    void setData(void *data) {_data = data;}
    void *data() {return _data;}
  private:
    void draw(QPaintDevice *dev, int xpos, int ypos);
    void setLevel(int val) {_level = val;}
    int level() {return _level;}
    QRect rect() {return _rect;}

    //! Flag for highlighting
    bool _highlighted;
    //! Flag for expanding
    bool _expanded;
    //! Indentation level for this node. Used by TreeNode::draw
    int _level;
    //! Pixmap data for this node.
    QPixmap *_pic;
    //! Text data for this node.
    QString _txt;
    //! Data associated to this node
    void *_data;
    //! The on-screen rectangle occupied by this node
    QRect _rect;

    friend class TreeWidget;
};

typedef QList<TreeNode> NodeList;
typedef QListIterator<TreeNode> NodeListIt;

/*!
  \class
    TreeWidget tree.h
  \brief
    The TreeWidget class provides a scrollable tree view a la Windows
    Explorer.

  The TreeWidget class provides an area on the screen which contains a
  hierarchy of nodes. Each node consists of a text and an optional pixmap.
  Each node can be associated with a piece of data which can be requested
  later (see \e setNodeData()).
  The TreeWidget uses a pixmap buffer to ensure smooth updates. It comes
  with two scrollbars for x and y direction which become active
  automatically.
  Nodes in the tree can be highlighted by pressing the left mouse button.
  The right mouse button is used to pop up optional menus. The user can
  provide these menus in the form of QPopupMenu pointers.
*/

class TreeWidget: public QFrame
{
  Q_OBJECT
  public:
    TreeWidget(QWidget *parent=0, const char *name=0);
    ~TreeWidget();
    void setCurrentNode(TreeNode *val);
    TreeNode *rootNode();
    TreeNode *currentNode();
    NodeList *nodes();
    TreeNode *addNode(const char *txt, QPixmap *pic=0, void *data=0,
                       TreeNode *parent=0);
    void addNode(TreeNode *id, TreeNode *parent=0);
    void rmNode(TreeNode *id);
    void clear();
    void setNodePopup(QPopupMenu *menu) {_nodePopup = menu;}
    void setBackPopup(QPopupMenu *menu) {_backPopup = menu;}
  public slots:
    void update();
  private slots:
    void doXBar(int val) {_offset.setX(val); update();}
    void doYBar(int val) {_offset.setY(val); update();}
  signals:
    void highlighted();
    void selected();
    void nothingSelected();
  private:
    void mousePressEvent(QMouseEvent *);
    void mouseDoubleClickEvent(QMouseEvent *);
    void paintEvent(QPaintEvent *);
    void resizeEvent(QResizeEvent *);
    void draw();
  private:
    //! The horizontal scrollbar
    QScrollBar *_xBar;
    //! The vertical scrollbar
    QScrollBar *_yBar;
    //! The root node of the tree
    TreeNode *_rootNode;
    //! The currently selected node
    TreeNode *_currNode;
    //! A list of nodes in the tree
    QList<TreeNode> _nodeList;
    //! The background buffer used for flicker-free updates
    QPixmap *_canvas;
    //! The offset for drawing the (clipped) tree
    QPoint _offset;
    //! The popup menu for the tree nodes
    QPopupMenu *_nodePopup;
    //! The popup menu for the tree background
    QPopupMenu *_backPopup;
};

#endif
